# -*- coding: utf-8 -*-
"""
    rpi2caster is a CAT (computer-aided typesetting) software
    for the Monotype composition caster (a hot-metal typesetting machine).

    This project uses a control interface for 31/32 solenoid valves
    and a machine cycle sensor. It can control a casting machine
    or a pneumatic paper tape perforator from the Monotype keyboard.

    The rpi2caster package consists of three main utilities:

        * casting (composition, material etc.),

        * caster/interface testing and diagnostics

    Machine control utility also serves as a diagnostic program
    for calibrating and testing the machine and control interface.

"""
from collections import OrderedDict
from contextlib import suppress
import os

import click

from . import ui

# initialize the configuration for rpi2caster
USER_DATA_DIR = click.get_app_dir('rpi2caster', force_posix=True, roaming=True)
with suppress(FileExistsError):
    os.mkdir(USER_DATA_DIR, mode=0o775)

# pica = .1667, US pica = .166, Didot = .1776, Fournier = .1628
# make it configurable in the future
ui.PICA = 0.1776


class CommandGroup(click.Group):
    """Click group which allows using abbreviated commands,
    and arranges them in the order they were defined."""
    def __init__(self, name=None, commands=None, **attrs):
        if commands is None:
            commands = OrderedDict()
        elif not isinstance(commands, OrderedDict):
            commands = OrderedDict(commands)
        click.Group.__init__(self, name=name, commands=commands, **attrs)

    def list_commands(self, ctx):
        """List command names as they are in commands dict."""
        return self.commands.keys()

    def get_command(self, ctx, cmd_name):
        """Try to get a command with given partial name;
        in case of multiple match, abort."""
        retval = click.Group.get_command(self, ctx, cmd_name)
        if retval is not None:
            return retval
        matches = [x for x in self.list_commands(ctx)
                   if x.startswith(cmd_name)]
        if not matches:
            return None
        elif len(matches) == 1:
            return click.Group.get_command(self, ctx, matches[0])
        ctx.fail('Too many matches: %s' % ', '.join(sorted(matches)))
        return None


@click.group(invoke_without_command=True, cls=CommandGroup, help=__doc__,
             context_settings=dict(help_option_names=['-h', '--help']))
@click.version_option(None, '--version', '-V')
@click.option('verbosity', '-v', count=True, default=0,
              help='verbose mode (count, default=0)')
@click.option('--address', '-a', help='address (default: 127.0.0.1)',
              default='127.0.0.1')
@click.option('--port', '-p', help='port (default: 23017)', type=int,
              default=23017)
@click.pass_context
def cli(ctx, verbosity, address, port):
    """decide whether to go to a subcommand or enter main menu"""
    ctx.obj = dict(address=address, port=port)
    ui.verbosity = verbosity
    if not ctx.invoked_subcommand:
        ctx.invoke(cast)


@cli.group(invoke_without_command=True, cls=CommandGroup,
           options_metavar='[-aph]', subcommand_metavar='[what]')
@click.pass_context
def cast(ctx):
    """Cast type with a Monotype caster.

    Casts composition, material for handsetting, QR codes.
    Can also cast a diecase proof.

    Can also be run in simulation mode without the actual caster."""
    address, port = ctx.obj.get('address'), ctx.obj.get('port')
    from .casting import Casting
    # replace the context object for the subcommands to see
    ctx.obj = Casting(address, port)
    if not ctx.invoked_subcommand:
        ctx.obj.main_menu()


@cast.command('ribbon', options_metavar='[-h]')
@click.argument('ribbon', metavar='[filename]', type=click.File())
@click.pass_obj
def cast_ribbon(casting, ribbon):
    """Cast composition from file or database."""
    casting.cast_file(ribbon)


@cast.command('xls', options_metavar='[-h]')
@click.argument('files', metavar='[filenames]',
                type=click.File(mode='rb'), nargs=-1)
@click.pass_obj
def cast_excel_file(casting, files):
    """Cast composition from file or database."""
    casting.cast_xls(files)


@cast.command('material', options_metavar='[-h]')
@click.pass_obj
def cast_handsetting_material(casting):
    """Cast founts, sorts and spaces/quads."""
    casting.cast_material()


@cast.command('qrcode', options_metavar='[-h]')
@click.pass_obj
def cast_qr_code(casting):
    """Generate and cast QR codes."""
    casting.cast_qr_code()


@cast.command('proof')
@click.pass_obj
def cast_diecase_proof(casting):
    """Cast a matrix case proof."""
    casting.diecase_proof()


@cast.command('test', options_metavar='[-h]')
@click.pass_obj
def test_machine(casting):
    """Monotype caster testing and diagnostics."""
    casting.diagnostics()


@cast.command('align', options_metavar='[-h]')
@click.pass_obj
def align_machine(casting):
    """Monotype caster testing and diagnostics."""
    casting.calibrate_machine()


@cli.command(options_metavar='[-ht]')
@click.option('--testing', '-t', is_flag=True, flag_value=True,
              help='use a unstable/development version instead of stable')
def update(testing):
    """Update the software."""
    # Upgrade routine
    dev_prompt = 'Testing version (newest features, but unstable)? '
    if ui.confirm('Update the software?', default=False):
        use_dev_version = testing or ui.confirm(dev_prompt, default=False)
        pre = '--pre' if use_dev_version else ''
        os.system('pip3 install {} --upgrade rpi2caster'.format(pre))


@cli.command()
def meow():
    "Easter egg."
    text = r"""
    ╔═══════════════════════════════════════════════════════════════════════╗
    ║                                   _                              _    ║
    ║    /\___/\       __   __     __   |    __  _ __       __   __    |    ║
    ║   |       |     /  \ /  \   /  \ -+-  |  \ |/  \     /  \ /  \  -+-   ║
    ║  _  *   *  _   |      __ |  \__   |   +--/ |        |      __ |  |    ║
    ║  -   /_\   -   |     /  \|     \  |   |    |        |     /  \|  |    ║
    ║      ---        \__/_\__/\__\__/___\_/ \__/|_        \__/_\__/\___\_  ║
    ║                                                                       ║
    ║ Hello Kitty!              gives your MONOTYPE nine lives              ║
    ║                                                                       ║
    ╚═══════════════════════════════════════════════════════════════════════╝
    """
    try:
        ui.display('\nOh, this was meowsome.\n')
        ui.display(text)
    except (OSError, ImportError, FileNotFoundError):
        print('There are no Easter Eggs in this program.')



# -*- coding: utf-8 -*-
"""Fundamental models for rpi2caster, not depending on database.
Classes-only module. All constants come from definitions."""

import json
from collections import namedtuple, OrderedDict
from pkg_resources import resource_string as rs

# some named tuples for easier attribute addressing
WedgeLimits = namedtuple('WedgeLimits', 'shrink stretch')
Matrix = namedtuple('Matrix',
                    'column row units row_units code '
                    'comment ribbon_entry wedges')
ParsedRecord = namedtuple('ParsedRecord',
                          ('raw signals comment column row has_signals is_char'
                           ' uses_row_16 has_0005 has_0075 has_s_needle '
                           'is_pump_start is_pump_stop is_newline'))

# constant wedge definitions data

# Wedge definitions: {WEDGE_SERIES: [row1_units, row2_units...]...}
WEDGES_JSON = rs('rpi2caster.data', 'wedge_units.json').decode()
WEDGE_DEFINITIONS = json.loads(WEDGES_JSON)
WEDGE_ALIASES = ('10E: UK S527', '10L: UK S536', '11Q: UK S574',
                 '14E: UK S1406', '1A: UK S207', '1B: UK S209',
                 '1C: UK S210', '1O: UK S221', '1R: UK S223',
                 '1Q: UK S224', '2A: UK S233', '2Z: UK S261',
                 '3O: UK S275', '3Q: UK S277', '3Y: UK S286',
                 '4A: UK S295', '4G: UK S300', '5P: UK S327',
                 '5V: UK S371', '7J: UK S422', '7Z: UK S449',
                 '8A: UK S450', '8U: UK S409', 'TW: S535 typewriter',
                 'AK: EU S5', 'BO: EU S221', 'CZ: EU S336',
                 'A: UK S5', 'D: UK S46', 'E: UK S92',
                 'F: UK S94', 'G: UK S93', 'J: UK S101',
                 'K: UK S87', 'L: UK S96', 'M: UK S45',
                 'N: UK S88', 'O: UK S111', 'Q: UK S50',
                 'S: UK S197', 'V: UK S202', 'W: UK S205', 'X: UK S47')


class Wedge:
    """Default S5-12E wedge, unless otherwise specified"""
    __slots__ = ('series', 'set_width', 'is_brit_pica', '_units')
    S5 = [5, 6, 7, 8, 9, 9, 9, 10, 10, 11, 12, 13, 14, 15, 18]
    definitions = WEDGE_DEFINITIONS
    aliases = WEDGE_ALIASES

    def __init__(self, wedge_name='', wedge_data=None):
        w_data = wedge_data or {}
        if wedge_name:
            self.name = wedge_name
        else:
            self.series = w_data.get('series', '5')
            self.set_width = w_data.get('set_width', 12.0)
            self.is_brit_pica = w_data.get('is_brit_pica', True)
            self._units = w_data.get('units', Wedge.S5)

    def __str__(self):
        return self.name

    def __repr__(self):
        return '<Wedge: {}>'.format(self.name)

    def __getitem__(self, row):
        return self.units[row]

    def __bool__(self):
        return True

    @property
    def pica(self):
        """Get the pica base of the wedge. Two values were used:
            .1667" (old British pica, European wedges), also: DTP pica,
            .1660" (new British, American), also: TeX pica."""
        return 0.1667 if self.is_brit_pica else 0.166

    @property
    def name(self):
        """Gets a wedge name - for example S5-12.25E.
        S (for stopbar) is prepended by convention.
        E is appended whenever the wedge is based on pica = .1667".
        """
        # Truncate the fractional part of the set width if it's an integer
        set_w = self.set_width if self.set_width % 1 else int(self.set_width)

        name = 'S{series}-{set_width}{european_suffix}'
        return name.format(series=self.series, set_width=set_w,
                           european_suffix=self.is_brit_pica * 'E')

    @name.setter
    def name(self, wedge_name):
        """Parse the wedge name to get series, set width, unit values
        and whether the wedge is British pica."""
        if not wedge_name:
            # use default S5
            return
        try:
            # For countries that use comma as decimal delimiter,
            # convert to point:
            wedge_name = str(wedge_name).replace(',', '.').upper().strip()
            # Check if this is an European wedge
            # (these were based on pica = .1667" )
            is_brit_pica = wedge_name.endswith('E')
            # Away with the initial S, final E and any spaces before and after
            # Make it work with space or dash as delimiter
            wedge = wedge_name.strip('SE ').replace('-', ' ').split(' ')
            series, raw_set_w = wedge
            # get the set width i.e. float approximated to .25
            set_w_str = ''.join(x for x in raw_set_w
                                if x.isnumeric() or x == '.')
            set_width = float(set_w_str) // 0.25 * 0.25
            if set_width > 25:
                raise ValueError
            # try to get the unit values, otherwise S5
            units = WEDGE_DEFINITIONS.get(series, Wedge.S5)
            # update the attributes
            self.series, self.set_width = series, set_width
            self.is_brit_pica, self.units = is_brit_pica, units
        except (TypeError, AttributeError, ValueError):
            # In case parsing goes wrong
            raise ValueError('Cannot parse wedge name {}'.format(wedge_name))

    @property
    def parameters(self):
        """Gets a list of parameters"""
        em_inches = round(self.pica * self.set_width / 12, 5)
        parameters = OrderedDict()
        parameters['Wedge designation'] = self.name
        parameters['Inch width of 18 units [1em]'] = em_inches
        parameters['Units per row'] = ' '.join(str(x) for x in self.units if x)
        return parameters

    @property
    def units(self):
        """Gets the unit values for the wedge's rows"""
        units = self._units or Wedge.S5
        # Add 0 at the beginning so that the list can be indexed with
        # row numbers - i.e. units[1] for row 1
        if units[0] is not 0:
            units = [0] + units
        # Ensure that wedge has 0 + 16 steps - i.e. length is 17
        # Keep adding the last value until we have enough
        while len(units) < 17:
            units.append(units[-1])
        return units

    @units.setter
    def units(self, units):
        """Sets the wedge unit values"""
        self._units = units

    def units_to_inches(self, units=1):
        """Convert units to inches, based on wedge's set width and pica def"""
        return round(units / 18 * self.set_width / 12 * self.pica, 4)

    def inches_to_units(self, inches=1):
        """Convert inches to units"""
        return round(18 * inches / self.pica * 12 / self.set_width, 2)

    def get_adjustment_limits(self):
        """Get the unit adjustment limits for this wedge.
        Without adjustment, the character has the same width as if the
        wedges were at 3/8.

        lower limit is reached at 1/1 0075 and 0005 wedge positions,
        max shrink is 3/8 - 1/1 = 2/7 => 2 * 0.0075 + 7 * 0.0005 = 0.0185"

        upper limit is reached at 15/15,
        max stretch is 15/15 - 3/8 = 12/7 => 12 * 0.0075 + 7 * 0.0005 = 0.0935"

        Constraints:
            Single-cell matrices have max .18" width for safety.
            Double-cell (large composition, wide characters) are .38"
            Otherwise the mould opening would not be covered by the matrix,
            leading to lead splashing over the diecase.

            Low spaces don't have this constraint as the upper mould blade
            prevents the lead from even reaching the diecase.

        Calculate this to wedge set units.
        Return positive integers.
        """
        # upper limit based on parameters
        stretch_inches = 0.0935
        cap = 0.18
        maximum = min(stretch_inches, cap)
        # calculate adjustment range in units
        shrink = int(18 * 12 * 0.0185 / self.pica / self.set_width)
        stretch = int(18 * 12 * maximum / self.pica / self.set_width)
        return WedgeLimits(shrink, stretch)

    def corrections(self, row, units):
        """Calculate the 0075 and 0005 wedge positions for this matrix
        based on the current wedge used.

        matrix - a Matrix object
        correction - units of self.wedge's set to add/subtract,
        units - arbitrary character width in units of self.wedge's set"""
        def steps(unit_width=0):
            """get a width (in .0005" steps) of a character
            for a given number of units or diecase row"""
            inches = unit_width / 18 * self.set_width / 12 * self.pica
            # return a number of 0.0005" steps
            return int(2000 * inches)

        def limits_exceeded():
            """raise an error if width can't be adjusted with wedges"""
            limits = self.get_adjustment_limits()
            minimum = row_units - limits.shrink
            maximum = row_units + limits.stretch
            message = ('{} units exceed adjustment limits (min: {} / max: {})')
            error_msg = message.format(units, minimum, maximum)
            raise ValueError(error_msg)

        # absolute width: how many .0005" steps is it?
        char_width = steps(units)
        # how wide would a character from the given row normally be?
        # (using self.wedge, not self.diecase.wedge!)
        row_units = self[row]
        row_width = steps(row_units)
        # calculate the difference and wedge positions
        # 1 step of 0075 wedge is 15 steps of 0005; neutral positions are 3/8
        # 3 * 15 + 8 = 53, so any increment/decrement is relative to this
        increments = char_width - row_width + 53
        # Upper limit: 15/15 => 15*15=225 + 15 = 240;
        # lower limit:  1/ 1 => 1 * 15 + 1 = 16
        if increments < 16 or increments > 240:
            limits_exceeded()
        # calculate wedge positions from the increments
        pos_0005, pos_0075 = increments % 15, increments // 15
        if not pos_0005:
            # wedge positions cannot be zero
            pos_0005, pos_0075 = 15, pos_0075 - 1
        return (pos_0075, pos_0005)


class Ribbon:
    """Ribbon model for the rpi2caster casting program"""

    def __init__(self, description='', diecase='', wedge='', contents=''):
        self.description = description
        self.diecase = diecase
        self.wedge = wedge
        self.contents = contents

    def __iter__(self):
        return iter(self.contents)

    def __next__(self):
        yield from self.contents

    def __repr__(self):
        return ('<Ribbon: {} diecase: {}, wedge: {}>'
                .format(self.description, self.diecase, self.wedge))

    def __bool__(self):
        return bool(self.contents)

    def __call__(self):
        return self

    @property
    def parameters(self):
        """Gets a list of parameters"""
        parameters = OrderedDict()
        parameters['Description'] = self.description
        parameters['Diecase'] = self.diecase
        parameters['Wedge'] = self.wedge
        return parameters



# -*- coding: utf-8 -*-
"""datatypes - custom data types and type handling functions for rpi2caster"""
from collections import Iterable, namedtuple, OrderedDict
from contextlib import suppress

# define aliases for boolean values
TRUE_ALIASES = ['true', 'on', 'yes', '1', 't', 'y']
FALSE_ALIASES = ['false', 'off', 'no', '0', 'f', 'n']

# start with an empty dictionary of types and limited parameters;
# after classes are defined, update it
TYPES = OrderedDict()
LIMITED_PARAMETERS = {len: 'length'}

# Type handling routines for common datatypes
# from_str: function converting a string to a value of this datatype,
# to_str: function transforming this datatype to string,
# validated_parameter : function(value) -> param. checked against limits
# type_name : string displayed in prompts, validation messages etc.,
# instancecheck: function used by the ValueListMetaclass
#       for checking the value type
#       isinstance(value, type) -> True if instancecheck returns True
Handler = namedtuple('TypeHandler',
                     ('from_str to_str validated_parameter '
                      'type_name instancecheck'))


def str_to_float(value):
    """Convert a string to float. Works with commas and dots as well."""
    # simplest approach first
    with suppress(ValueError):
        return float(value)

    # maybe it has a comma instead of dot?
    # (some locales use commas as decimal points)
    with suppress(AttributeError, ValueError):
        replaced_comma = value.replace(',', '.')
        return float(replaced_comma)

    # binary, hexadecimal and octal strings should be converted to int
    # then convert to float
    with suppress(ValueError):
        integer = int(value, base=0)
        return float(integer)

    # nothing else to do...
    raise TypeError('{} is not a valid real number'.format(value))


def str_to_int(value):
    """Convert a string to int.
    Floats will be rounded according to math rules.
    Hexstrings and binary strings will be converted as well."""
    # simplest approach first
    with suppress(ValueError):
        return int(value)

    # it may be a binary string or hexstring...
    with suppress(ValueError):
        return int(value, base=0)

    # it may be a float
    with suppress(AttributeError, ValueError):
        float_value = str_to_float(value)
        return round(float_value)

    # nothing else to do...
    raise TypeError('{} is not a valid integer number'.format(value))


def str_to_bool(value=None):
    """Convert a string to boolean.
    0, no, false, off, n, f, none, empty iterable, None, whitespace -> False
    1, yes, true, on, y, t -> True
    any other value -> None

    Failed conversion raises a TypeError with user-friendly message.
    """
    err = 'Invalid value: "{value}". Correct values: {correct}.'
    correct = TYPES[bool].type_name
    values = dict()
    values.update({name: True for name in TRUE_ALIASES})
    values.update({name: False for name in FALSE_ALIASES})
    # try the simplest approach first...
    if not value:
        return False

    # non-empty string/list/etc.? make a string and normalize it
    value_string = str(value).strip().lower()
    if not value_string:
        return False

    # look it up in valid values
    retval = values.get(value_string)

    # unknown values raise ValueError
    if retval is None:
        raise TypeError(err.format(value=value, correct=correct))

    # return True or False
    return retval


def get_type(value):
    """Determine a type of the value. If it fails, default to str."""
    # implementing this with singledispatch failed:
    # custom datatypes were not recognized
    if value is None:
        return None

    # iterate until a match is found
    for datatype, handler in TYPES.items():
        with suppress(TypeError):
            if handler.instancecheck(value):
                return datatype

    return str


def get_handler(datatype):
    """Get a type handler; unknown types will default to str."""
    return TYPES.get(datatype) or TYPES.get(str)


def get_string(value=None, datatype=str):
    """Get a printable string of a value that would act as a prefill
    value for input prompts."""
    to_str = get_handler(datatype).to_str
    return to_str('' if value is None else value)


def get_false_value(datatype):
    """Returns a value (of desired datatype) evaluating False."""
    for candidate in ('', [], (), 0, False):
        with suppress(TypeError, ValueError):
            ret = datatype(candidate)
            if not ret:
                # found a match!
                return ret

    # could not find a matching datatype value...
    return None


def convert_and_validate(value, default=None, datatype=None,
                         minimum=None, maximum=None,
                         condition=lambda x: True):
    """Wrap a function outputting a string to convert it,
    based on default value or explicitly specified type:

    If conversion fails, raises TypeError.
    If validation fails, raises ValueError.
    If no input and default is None, raises TypeError.

    value: a value to be checked,

    default: a value, exception or None.

             Empty input will result in raising an exception
             (if default is Exception subclass or its instance),
             or returning default value (coerced to a specified datatype).

    datatype: explicitly specified type.
              If None, default value's datatype will be used if possible;
              in case of exceptions etc. the function defaults to str.

    minimum, maximum : limits applying to validated parameter (length or value)
                       of the return value's datatype,

    condition: additional condition for validation to pass.
    """
    # desired type:
    # specified datatype -> type of default value -> string
    default_value_type = get_type(default)
    retval_datatype = datatype or default_value_type or str

    # no input? then raise any exceptions supplied in default value,
    # or return the default value itself
    if not value:
        # None is a special case, return it right away
        if default is None:
            raise ValueError('A value is required.')

        # return a default value then
        default_retval = try_raising(default)
        return default_retval or get_false_value(retval_datatype)

    # desired datatype handler provides info about what to validate
    retval_type_handler = get_handler(retval_datatype)
    validated_parameter = retval_type_handler.validated_parameter

    # convert, validate, return
    converted = convert(value, retval_datatype)
    validated = validate(converted,
                         validated_parameter, minimum, maximum, condition)
    return validated


def try_raising(value):
    """Try to raise an exception from option in the first place.
    Raises from classes (e.g. ValueError)
    or exception objects (e.g. ValueError('Incorrect value!'))

    If that fails, returns the value itself.
    """
    # raise TypeError classes and instances
    if value is TypeError or isinstance(value, TypeError):
        raise value

    # try to raise other exceptions
    try:
        raise value
    except TypeError:
        # it was not an exception after all
        return value


def itself(value):
    """Do-nothing function which returns the input value;
    used as an no-op implementation of converter for types that don't need it,
    equivalent to lambda value: value"""
    return value


def convert(value='', desired_type=str):
    """Converts a value to desired type
    and raises TypeError in case ValueError would normally be raised
    (avoiding TypeError in some cases and ValueError in others,
    e.g. converting a non-numeric string to int).

    Failed conversion raises a TypeError with a user-friendly message.
    """
    # maybe conversion is not necessary at all...
    source_value_type = get_type(value)
    if source_value_type == desired_type:
        return value

    # convert the source value to a string if it's not it
    source_type_handler = get_handler(source_value_type)
    source_to_string = source_type_handler.to_str
    value_string = source_to_string(value)

    # is an explanation available for the desired type?
    # look it up (not found -> None -> AttributeError -> generic message)
    try:
        type_definition = TYPES.get(desired_type)
        description = type_definition.type_name
        converter = type_definition.from_str
        exc_message = 'Wrong value type. Correct type: {}'.format(description)
    except AttributeError:
        exc_message = 'Wrong value type.'
        converter = desired_type

    # normally everything should be okay...
    # otherwise raise with a user-friendly message
    try:
        converted_value = converter(value_string)
        return converted_value
    except TypeError:
        raise TypeError(exc_message)
    except ValueError:
        raise TypeError(exc_message)


def validate(value, validated_parameter=None,
             minimum=None, maximum=None, condition=None):
    """Validate the value.

    Check condition, if specified. If it evaluates True, process further.

    If validated parameter and minimum and/or maximum is given, check limits.

    If any check evaluates False, raise ValueError and inform what went wrong.

    value : value to be checked,
    validated_parameter : fun(x)->limited parameter, or None (skip limit check)
                          (examples: lambda x: x; len)

    minimum, maximum : min and max value of the validated parameter for the
                       validation to pass, if specified.
                       If None, don't check this limit.

    condition : fun(x)->bool, or None (skip condition check)
                If retval evaluates False, fail validation.
    """
    # parameter name: if available, provide more info for the user
    try:
        name = LIMITED_PARAMETERS.get(validated_parameter)
        message = 'current {value_name} of {value} is {issue} than {limit}.'
    except AttributeError:
        name = ''
        message = 'value is {issue} than {limit}.'

    # check the value against condition
    if condition is not None and not condition(value):
        raise ValueError('value does not match conditions')

    # is there anything to check? if not, pass right away
    if not validated_parameter:
        return value

    # then check limits
    parameter_value = validated_parameter(value)

    if minimum is not None and parameter_value < minimum:
        # validation error
        error_data = dict(value_name=name, value=parameter_value,
                          issue='smaller', limit=minimum)
        raise ValueError(message.format(**error_data))

    if maximum is not None and parameter_value > maximum:
        # validation error
        error_data = dict(value_name=name, value=parameter_value,
                          issue='greater', limit=maximum)
        raise ValueError(message.format(**error_data))

    # validation pass
    return value


def make_handler(outer_type, inner_type):
    """Make a handler for container type for specified inner types"""
    container = Container(outer_type, inner_type)
    type_names = {int: 'integer numbers', float: 'decimal numbers'}
    name = type_names.get(inner_type, 'values')
    return Handler(from_str=container,
                   to_str=lambda v: ', '.join(str(i) for i in v),
                   validated_parameter=len,
                   instancecheck=container.instancecheck,
                   type_name='comma-separated {}'.format(name))


class Container:
    """A container (tuple, list) of specified inner types"""
    def __init__(self, outer_type, inner_type):
        self.outer_type = outer_type
        self.inner_type = inner_type

    def __call__(self, source):
        items = self.parse(source)
        return self.convert(items)

    @staticmethod
    def parse(source):
        """check and dispatch on inner value type"""
        if isinstance(source, str):
            items = (x.strip() for x in source.split(','))
        elif isinstance(source, Iterable):
            items = iter(source)
        else:
            # single item -> make a tuple
            items = (source, )

        return items

    def convert(self, items):
        """convert the values to a desired type"""
        collection = (convert(x, self.inner_type) for x in items)
        return self.outer_type(collection)

    def instancecheck(self, value):
        """check if we have the instance of this container"""
        checks = {int: lambda x: all(isinstance(v, int) for v in x),
                  float: lambda x: all(isinstance(v, (int, float)) for v in x),
                  str: ' '.join}
        try:
            check_inner = checks.get(self.inner_type)
            return isinstance(value, self.outer_type) and check_inner(value)
        except (TypeError, ValueError):
            return False


# itself function is defined => create a limit name for it
LIMITED_PARAMETERS[itself] = 'value'
# define datatype handlers
# and add them in the sequence from most to least specific
# (type checking is iterative and we don't want generic types like list
# to overshadow specific ones like list of integers)
TYPES[bool] = Handler(from_str=str_to_bool, to_str=str,
                      validated_parameter=None,
                      type_name='y/n, yes/no, on/off,  1/0, true/false, t/f',
                      instancecheck=lambda i: i is True or i is False)

TYPES[Exception] = Handler(from_str=itself, to_str=lambda _: '',
                           validated_parameter=len,
                           type_name='',
                           # is an exception object or Exception subclass
                           instancecheck=lambda i:
                           isinstance(i, Exception) or
                           isinstance(i, type) and issubclass(i, Exception))

TYPES[int] = Handler(from_str=str_to_int,
                     to_str=str,
                     validated_parameter=itself,
                     type_name='integer number e.g. 5',
                     instancecheck=lambda i: isinstance(i, int))

TYPES[float] = Handler(from_str=str_to_float,
                       to_str=lambda x: str(float(x)),
                       validated_parameter=itself,
                       type_name='decimal number e.g. 5.5',
                       instancecheck=lambda i: isinstance(i, float))

TYPES[str] = Handler(from_str=str, to_str=str,
                     validated_parameter=len,
                     type_name='',
                     instancecheck=lambda i: isinstance(i, str))

# add handlers for container types
for outer in (list, tuple, set):
    for inner in (int, float, str):
        TYPES[(outer, inner)] = make_handler(outer, inner)
    # add a generic list/tuple/set not specifying inner type
    TYPES[outer] = make_handler(outer, str)



# -*- coding: utf-8 -*-
"""Caster object for either real or virtual Monotype composition caster"""
# Standard library imports
from collections import deque, OrderedDict
from functools import wraps
from json.decoder import JSONDecodeError
import time

import requests
import librpi2caster

# Intra-package imports
from . import ui, stats
from .functions import parse_record, parse_signals

ON, OFF, HMN, KMN, UNITSHIFT = True, False, 'HMN', 'KMN', 'unit shift'


def caster_factory(address, port):
    """Make a single instance of MonotypeCaster for a given address and port
    when needed. After it's there, return it when calling the function
    with this address and port.

    None, None denote a simulation caster."""
    try:
        return MonotypeCaster.instances[(address, port)]
    except KeyError:
        return MonotypeCaster(address, port)


def handle_communication_error(routine):
    """If something goes wrong with a connection, ask what to do."""
    @wraps(routine)
    def wrapper(*args, **kwargs):
        """wrapper function"""
        try:
            return routine(*args, **kwargs)
        except librpi2caster.CommunicationError as error:
            message = ('Connection to the interface lost. Check your link.\n'
                       'Original error message:\n{}\n'
                       'Do you want to try again, or abort?')
            if ui.confirm(message.format(error),
                          abort_answer=False, allow_abort=False):
                return routine(*args, **kwargs)
    return wrapper


class SimulationCaster:
    """Common methods for a caster class."""
    def __init__(self, address='127.0.0.1', port=23017):
        self.url = 'http://{}:{}'.format(address, port)
        self.row16_mode = OFF
        self.config = dict(punching_on_time=0.2, punching_off_time=0.3,
                           sensor_timeout=5, pump_stop_timeout=20,
                           startup_timeout=20, name='Simulation interface',
                           punch_mode=False)
        self.status = dict(water=OFF, air=OFF, motor=OFF, pump=OFF,
                           wedge_0005=15, wedge_0075=15, speed='0rpm',
                           machine=OFF, signals=[])
        self.update_config()
        self.update_status()

    def __str__(self):
        return self.config['name']

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, *_):
        self.stop()

    @property
    def punch_mode(self):
        """Checks if the interface is in ribbon punching mode."""
        return self.config.get('punch_mode')

    @punch_mode.setter
    def punch_mode(self, state):
        """Sets the ribbon punching mode."""
        self.update_config(punch_mode=bool(state))

    @property
    def testing_mode(self):
        """Checks the testing mode where signals are advanced manually."""
        return self.status.get('testing_mode')

    @testing_mode.setter
    def testing_mode(self, state):
        """Sets the testing mode"""
        self.update_status(testing_mode=bool(state))

    @property
    def parameters(self):
        """Gets a list of parameters"""
        parameters = OrderedDict({'': 'Caster parameters'})
        parameters.update(**self.config)
        return parameters

    @property
    def emergency_stop(self):
        """Check the emergency stop state"""
        return self.status.get('emergency_stop')

    @emergency_stop.setter
    def emergency_stop(self, state):
        """Activate or reset the machine emergency stop"""
        self.status.update(emergency_stop=state)
        if state:
            self.stop()
            raise librpi2caster.MachineStopped

    def update_status(self, **kwargs):
        """Update the machine status."""
        self.status.update(**kwargs)

    def update_config(self, **kwargs):
        """Update the configuration."""
        self.config.update(**kwargs)

    def choose_row16_mode(self, row16_needed=False):
        """Choose the addressing system for the diecase row 16."""
        if row16_needed and self.row16_mode:
            ui.display('Keep the {} attachment ON.'.format(self.row16_mode))
        elif row16_needed and not self.row16_mode:
            # choose the row 16 addressing mode
            prompt = 'Choose the row 16 addressing mode'
            options = [ui.option(key=name[0], value=name, text=name)
                       for name in [HMN, KMN, UNITSHIFT]]
            options.append(ui.option(key='o', value=OFF,
                                     text='Off - cast from row 15 instead'))
            if len(options) > 1:
                self.row16_mode = ui.simple_menu(prompt, options,
                                                 default_key='o',
                                                 allow_abort=True)
            else:
                self.row16_mode = OFF

        elif self.row16_mode and not row16_needed:
            # turn off the unneeded row 16 adddressing attachment
            ui.display('The {} attachment is not needed - turn it off.'
                       .format(self.row16_mode))
            self.row16_mode = OFF

        else:
            ui.display('No row 16 addressing attachment is needed.')

    @property
    def signals(self):
        """Get the signals from the machine"""
        return self.status.get('signals', [])

    def start(self):
        """Simulates the machine start"""
        if self.emergency_stop:
            if ui.confirm('\n**********\nWARNING!!!\n**********\n'
                          '\nEmergency stop is engaged!\n'
                          'It is necessary to clear it before starting.\n'
                          'Answering "no" will abort casting.',
                          default=None, abort_answer=False, allow_abort=False):
                self.emergency_stop = OFF
        # normal starting sequence
        self.update_status(air=ON)
        if not any((self.punch_mode, self.testing_mode)):
            self.update_status(water=ON, motor=ON)
        self.update_status(machine=ON)

    def stop(self):
        """Simulates the machine stop"""
        if not any((self.punch_mode, self.testing_mode)):
            self.update_status(pump=OFF, water=OFF, motor=OFF)
        self.update_status(air=OFF, machine=OFF, testing_mode=OFF)

    def send(self, signals, timeout=None, request_timeout=None):
        """Simulates sending the signals to the caster."""
        def update_wedges_and_pump():
            """Check the pump state and wedge positions;
            update the current status if neeeded."""
            def found(code):
                """check if code was found in a combination"""
                return set(code).issubset(signals)

            # check the previous wedge positions and pump state
            pos_0075 = self.status.get('wedge_0075')
            pos_0005 = self.status.get('wedge_0005')
            pump_working = self.status.get('pump')
            # check 0005 wedge position:
            # find the earliest row number or default to 15
            if found(['0005']) or found('NJ'):
                pump_working = OFF
                for pos in range(1, 15):
                    if str(pos) in signals_sent:
                        pos_0005 = pos
                        break
                else:
                    pos_0005 = 15

            # check 0075 wedge position and determine the pump status:
            # find the earliest row number or default to 15
            if found(['0075']) or found('NK'):
                # 0075 always turns the pump on
                pump_working = ON
                for pos in range(1, 15):
                    if str(pos) in signals_sent:
                        pos_0075 = pos
                        break
                else:
                    pos_0075 = 15

            self.update_status(pump=bool(pump_working),
                               wedge_0075=pos_0075, wedge_0005=pos_0005)

        start_time = time.time()
        max_wait_time = max(timeout or 0, request_timeout or 0) or 10
        # convert signals based on modes
        codes = parse_signals(signals, self.row16_mode)
        signals_string = ''.join(codes)
        try:
            if self.testing_mode:
                ui.display('Testing signals: {}'.format(signals_string))
                signals_sent = codes
            elif self.punch_mode:
                signals_sent = (codes if len(codes) >= 2
                                else codes if 'O15' in codes
                                else [*codes, 'O15'])
                ui.display('punches going up')
                time.sleep(self.config['punching_on_time'])
                ui.display('punches going down')
                time.sleep(self.config['punching_off_time'])
            else:
                signals_sent = [s for s in codes if s != 'O15']
                ui.display('photocell ON')
                time.sleep(0.5)
                ui.display('photocell OFF')
                time.sleep(0.5)
            self.update_status(signals=signals_sent)
            update_wedges_and_pump()
            if (time.time() - start_time) > max_wait_time:
                raise librpi2caster.MachineStopped
        except (KeyboardInterrupt, EOFError):
            self.emergency_stop = ON
        except librpi2caster.InterfaceNotStarted:
            self.start()
        finally:
            self.update_status()

    def cast_one(self, record, timeout=None):
        """Casting sequence: sensor on - valves on - sensor off - valves off"""
        try:
            signals = record.signals
        except AttributeError:
            signals = record
        # calculate the timeout to prevent the request
        # from hanging indefinitely
        if timeout:
            request_timeout = 2 * timeout + 5
        else:
            request_timeout = 2 * self.config['sensor_timeout'] + 2
        self.send(signals, timeout=timeout, request_timeout=request_timeout)

    def punch_one(self, signals):
        """Punch a single signals combination"""
        off_time = self.config['punching_off_time']
        on_time = self.config['punching_on_time']
        timeout = on_time + off_time + 5
        self.send(signals, request_timeout=timeout)

    def test_one(self, signals):
        """Test a single signals combination"""
        self.testing_mode = True
        self.send(signals, request_timeout=5)

    def cast_or_punch(self, ribbon):
        """Cast / punch a ribbon, based on the selected operation mode."""
        return (self.punch(ribbon) if self.punch_mode
                else self.advanced_cast(ribbon))

    def simple_cast(self, input_sequence, ask=True, repetitions=1):
        """Cast a series of multiple records.
        This is a simplified routine without statistics."""
        source = [parse_record(item) for item in input_sequence]
        # do we need to cast from row 16 and choose an attachment?
        row16_in_use = any((record.uses_row_16 for record in source))
        self.choose_row16_mode(row16_in_use)
        # use caster context to check machine rotation and ensure
        # that no valves stay open after we're done
        with self:
            # cast as many as initially ordered and ask to continue
            while repetitions > 0:
                try:
                    for record in source:
                        found = '{r.signals:<20}{r.comment}'.format(r=record)
                        ui.display(found)
                        self.cast_one(record)
                        signals = self.status.get('signals', [])
                        ui.display('casting: {}'.format(' '.join(signals)))
                    # repetition successful
                    repetitions -= 1
                except librpi2caster.MachineStopped:
                    # repetition failed
                    if ui.confirm('Machine stopped: continue casting?',
                                  abort_answer=False):
                        # restart the machine
                        self.start()
                if ask and not repetitions and ui.confirm('Repeat?'):
                    repetitions += 1

    def advanced_cast(self, input_sequence):
        """Main casting routine.

        First check if ribbon needs rewinding (when it starts with pump stop).

        Ask user about number of repetitions (useful for e.g. business cards
        or souvenir lines), number of initial lines skipped for all runs
        and the upcoming run only.

        Ask user whether to pre-heat the mould
        to stabilize the temperature and cast good quality type.

        Cast the ribbon single or multiple times, displaying the statistics
        about current record and casting progress.

        If casting multiple runs, repeat until all are done,
        offer adding some more.
        """
        def rewind_if_needed():
            """Decide whether to rewind the ribbon or not.
            If casting and stop comes first, rewind. Otherwise not."""
            nonlocal ribbon
            for record in ribbon:
                if record.is_pump_stop:
                    # rewind
                    ribbon = [x for x in reversed(ribbon)]
                    return
                if record.is_pump_start:
                    # no need to rewind
                    return

        def skip_lines(source):
            """Skip a definite number of lines"""
            # Apply constraints: 0 <= lines_skipped < lines in ribbon
            lines_skipped = stats.run_lines_skipped()
            if lines_skipped:
                ui.display('Skipping {} lines'.format(lines_skipped))
            # Take away combinations until we skip the desired number of lines
            # BEWARE: ribbon starts with galley trip!
            # We must give it back after lines are taken away
            record = parse_record('')
            sequence = deque(source)
            while lines_skipped > 0:
                record = sequence.popleft()
                lines_skipped -= 1 * record.is_newline
            # give the last code back
            sequence.appendleft(record)
            return sequence

        def set_lines_skipped(run=True, session=True):
            """Set the number of lines skipped for run and session
            (in case of multi-session runs)."""
            # allow skipping only if ribbon is more than 2 lines long
            limit = max(0, stats.ribbon_lines() - 2)
            if not limit:
                return
            # how many can we skip?
            if run or session:
                ui.display('We can skip up to {} lines.'.format(limit))

            if run:
                # run lines skipping
                # how many lines were successfully cast?
                lines_done = stats.lines_done()
                if lines_done:
                    ui.display('{} lines were cast in the last run.'
                               .format(lines_done))
                # Ask user how many to skip (default: all successfully cast)
                r_skip = ui.enter('How many lines to skip for THIS run?',
                                  default=lines_done, minimum=0, maximum=limit)
                stats.update(run_line_skip=r_skip)

            if session:
                # Skip lines effective for ALL runs
                # session line skipping affects multi-run sessions only
                # don't do it for single-run sessions
                if stats.runs() > 1:
                    s_skip = ui.enter('How many lines to skip for ALL runs?',
                                      default=0, minimum=0, maximum=limit)
                    stats.update(session_line_skip=s_skip)

        def cast_queue(queue):
            """Casts the sequence of codes in given sequence.
            This function is executed until the sequence is exhausted
            or casting is stopped by machine or user."""
            # in punching mode, lack of row will trigger signal 15,
            # lack of column will trigger signal O
            # in punching and testing mode, signal O or 15 will be present
            # in the output combination as O15
            for record in queue:
                # check if signal will be cast at all
                if not record.has_signals:
                    ui.display_header(record.comment)
                    continue
                # display some info and cast the signals
                stats.update(record=record)
                self.cast_one(record)
                # prepare data for display
                info = OrderedDict()
                info['Signals sent'] = ' '.join(self.status.get('signals', []))
                info['Pump'] = 'ON' if self.status['pump'] else 'OFF'
                info['Wedge 0005 at'] = self.status['wedge_0005']
                info['Wedge 0075 at'] = self.status['wedge_0075']
                info['Speed'] = self.status['speed']
                # display status
                ui.display('*' * 80)
                ui.display_parameters(stats.code_parameters(), info)
                ui.display('*' * 80)

        # Ribbon pre-processing and casting parameters setup
        ribbon = [parse_record(code) for code in input_sequence]
        rewind_if_needed()
        # initialize statistics
        stats.reset()
        stats.update(ribbon=ribbon)
        # display some info for the user
        ui.display_parameters(stats.ribbon_parameters())
        # set the number of casting runs
        stats.update(runs=ui.enter('How many times do you want to cast this?',
                                   default=1, minimum=0))
        # initial line skipping
        set_lines_skipped(run=True, session=True)
        ui.display_parameters(stats.session_parameters())
        # check if the row 16 addressing will be used;
        # connect with the interface
        row16_in_use = any(record.uses_row_16 for record in ribbon)
        self.choose_row16_mode(row16_in_use)
        with self:
            while stats.runs_left():
                # Prepare the ribbon ad hoc
                queue = skip_lines(ribbon)
                stats.update(queue=queue)
                # Cast the run and check if it was successful
                try:
                    cast_queue(queue)
                    stats.update(casting_success=True)
                    if not stats.runs_left():
                        # user might want to re-run this
                        msg = 'Casting successfully finished. Any more runs?'
                        stats.update(runs=ui.enter(msg, default=0, minimum=0))

                except librpi2caster.MachineStopped:
                    stats.update(casting_success=False)
                    ui.display('The composition caster stopped working!\n')
                    # aborted - ask if user wants to continue
                    runs_left = stats.runs_left()
                    if runs_left > 1:
                        ui.confirm('{} runs left, continue?'.format(runs_left),
                                   default=True, abort_answer=False,
                                   allow_abort=False)
                    else:
                        ui.confirm('Do you want to repeat the casting job?',
                                   default=True, abort_answer=False,
                                   allow_abort=False)
                    # offer to skip lines for re-casting the failed run
                    skip_successful = stats.lines_done() >= 2
                    set_lines_skipped(run=skip_successful, session=False)
                    # restart the machine after emergency stop
                    self.start()

    def punch(self, signals_sequence):
        """Punching sequence: valves on - wait - valves off- wait"""
        # do we need to use any row16 addressing attachment?
        source = [parse_record(item) for item in signals_sequence]
        row16_in_use = any(record.uses_row_16 for record in source)
        self.choose_row16_mode(row16_in_use)
        # start punching
        with self:
            for record in source:
                try:
                    ui.display('{r.signals:<20}{r.comment}'.format(r=record))
                    self.punch_one(record.signals)
                    signals = self.status.get('signals', [])
                    ui.display('punching: {}'.format(' '.join(signals)))
                except librpi2caster.MachineStopped:
                    if ui.confirm('Machine stopped - continue?'):
                        self.punch_one(record.signals)
                        signals = self.status.get('signals', [])
                        ui.display('punching: {}'.format(' '.join(signals)))
                    else:
                        break
            else:
                ui.display('All codes successfully punched.')
            ui.pause('Punching finished. Take the ribbon off the tower.')

    def test(self, signals_sequence, duration=None):
        """Testing: advance manually or automatically"""
        source = [parse_record(item) for item in signals_sequence]
        self.testing_mode = True
        with self:
            while True:
                try:
                    pairs = zip(source, [*source[1:], None])
                    for record, next_record in pairs:
                        self.test_one(record.signals)
                        signals = self.status.get('signals', [])
                        ui.display('testing: {}'.format(' '.join(signals)))
                        if duration:
                            time.sleep(duration)
                        elif next_record:
                            ui.pause('Next: {}'.format(next_record.signals),
                                     allow_abort=True)
                except librpi2caster.MachineStopped:
                    ui.display('Testing stopped.')
                if len(signals_sequence) == 1:
                    ui.pause('Sending signals...')
                    break
                else:
                    ui.confirm('Repeat?', abort_answer=False)


class MonotypeCaster(SimulationCaster):
    """Methods common for Caster classes, used for instantiating
    caster driver objects (whether real hardware or mockup for simulation)."""
    instances = {(None, None): SimulationCaster()}

    def _request(self, path='', method='get', request_timeout=None, **kwargs):
        """Encode data with JSON and send it to self.url in a request."""
        errors = {exc.code: exc
                  for exc in (librpi2caster.MachineStopped,
                              librpi2caster.UnsupportedMode,
                              librpi2caster.UnsupportedRow16Mode,
                              librpi2caster.InterfaceBusy,
                              librpi2caster.InterfaceNotStarted)}
        url = '{}/{}'.format(self.url, path).strip('/')
        data = kwargs or {}
        try:
            # get the info from the server
            response = requests.request(method, url,
                                        json=data, timeout=request_timeout)
            # raise any HTTP errors ASAP
            response.raise_for_status()
            # we're sure we have a proper 200 OK status code
            reply = response.json()
            if not reply.get('success'):
                # shit happened in the driver
                error_code = reply.get('error_code', 666)
                exception = errors.get(error_code)
                if exception:
                    raise exception
            # return the content without the success value
            reply.pop('success')
            return reply
        except requests.exceptions.InvalidSchema:
            # wrong URL
            msg = 'The URL: {} must be a http://... or https://... address.'
            raise librpi2caster.ConfigurationError(msg.format(url))
        except requests.HTTPError as error:
            if response.status_code == 501:
                raise NotImplementedError('{}: feature not supported by server'
                                          .format(url))
            # 400, 404, 503 etc.
            raise librpi2caster.CommunicationError(message=str(error))
        except requests.Timeout:
            msg = 'Connection to {} timed out.'
            raise librpi2caster.CommunicationError(message=msg.format(url))
        except JSONDecodeError:
            msg = 'Connection to {} returned incorrect data (expected: JSON).'
            raise librpi2caster.CommunicationError(message=msg.format(url))
        except requests.ConnectionError:
            msg = 'Error connecting to {}.'
            raise librpi2caster.CommunicationError(message=msg.format(url))

    @handle_communication_error
    def send(self, signals='', timeout=None, request_timeout=None):
        """Send signals to the interface. Wait for an OK response."""
        try:
            codes = parse_signals(signals, self.row16_mode)
            data = dict(signals=codes, timeout=timeout)
            self._request('signals', method='post',
                          request_timeout=request_timeout, **data)
        except librpi2caster.InterfaceNotStarted:
            self.start()
            self._request('signals', method='post',
                          request_timeout=request_timeout, **data)
        except KeyboardInterrupt:
            self.emergency_stop = ON
            raise librpi2caster.MachineStopped
        except (EOFError, librpi2caster.CommunicationError):
            self.stop()
            raise librpi2caster.MachineStopped
        self.update_status()

    @handle_communication_error
    def start(self):
        """Machine startup sequence.
        In the casting mode:
            The interface will start the subsystems, if possible:
                compressed air supply, cooling water supply and motor.
            The operator has to turn the machine's shaft clutch on.
            The interface driver will detect rotation, and if the machine
            is stalling, a MachineStopped exception is raised.

        In other modes (testing, punching, manual punching):
            The interface will just turn on the compressed air supply.
        """

        try:
            if self.emergency_stop and ui.confirm(
                    '\n**********\nWARNING!!!\n**********\n'
                    '\nEmergency stop is engaged!\n'
                    'It is necessary to clear it before starting.\n'
                    'Answering "no" will abort casting.',
                    default=None, abort_answer=False, allow_abort=False):
                self.emergency_stop = OFF

            if self.testing_mode:
                request_timeout = 5
            elif self.punch_mode:
                ui.pause('Waiting for you to start punching...')
                request_timeout = 5
            else:
                info = ('Starting the composition caster now...\n\n'
                        'Turn on the motor if needed, and engage the clutch.\n'
                        'When the machine is turning, casting will begin.\n'
                        '\n\nCommence casting? (N = abort)')
                if ui.confirm(info, abort_answer=False, default=True):
                    ui.display('\nWait for a few turns of the cam shaft...\n')
                request_timeout = 3 * self.config['startup_timeout'] + 2
            # send the request and handle any exceptions
            try:
                self._request('machine', method='put',
                              request_timeout=request_timeout)
                if not self.punch_mode and not self.testing_mode:
                    ui.display('\nOK, the machine is running...\n')
            except librpi2caster.InterfaceBusy:
                if ui.confirm('\nThis interface is already busy.\n'
                              'If this is not the case, answer YES to '
                              'stop and reset it. Otherwise abort.\n',
                              abort_answer=False, default=None):
                    self.stop()
                    self.start()
            except KeyboardInterrupt:
                self.emergency_stop = ON
                self.start()
        except librpi2caster.MachineStopped:
            ui.display('\nThe machine was stopped because it was stalling,\n'
                       'or because emergency stop button was pressed.\n')
            self.start()
        finally:
            self.update_status()

    @handle_communication_error
    def stop(self):
        """Machine stop sequence.
        The interface driver checks if the pump is active
        and turns it off if necessary.

        In the casting mode, the driver will turn off the motor
        and cut off the cooling water supply.
        Then, the air supply is cut off.
        """
        try:
            if self.testing_mode or self.punch_mode:
                request_timeout = 3
            else:
                ui.display('Turning the machine off...')
                request_timeout = self.config['pump_stop_timeout'] * 2 + 2
            try:
                self._request('machine', method='delete',
                              request_timeout=request_timeout)
            except KeyboardInterrupt:
                self.emergency_stop = ON
        except librpi2caster.MachineStopped:
            self.stop()
        finally:
            self.update_status()

    @property
    def emergency_stop(self):
        """Check the emergency stop state"""
        self.update_status()
        return self.status.get('emergency_stop')

    @emergency_stop.setter
    @handle_communication_error
    def emergency_stop(self, state):
        """Activate or reset the emergency stop on the machine"""
        method = 'put' if state else 'delete'
        if state:
            ui.display('\n**********\nWARNING!!!\n**********\n'
                       'Emergency stop - stopping the machine NOW.\n\n'
                       'Turn the machine by hand a few times.\n')
        try:
            # will raise MachineStopped
            self._request('emergency_stop', method=method,
                          request_timeout=(120, 120))
        finally:
            self.update_status()

    @handle_communication_error
    def update_status(self, **kwargs):
        """Send the new (optional) parameters to the interface server.
        Update the local status afterwards."""
        status = self._request(method='post', request_timeout=(3, 3), **kwargs)
        self.status.update(status)

    @handle_communication_error
    def update_config(self, **kwargs):
        """Send the new (optional) parameters to the interface server.
        Update the local configuration afterwards."""
        config = self._request('config', method='post',
                               request_timeout=(3, 3), **kwargs)
        self.config.update(config)



# -*- coding: utf-8 -*-
"""Casting utility: cast or punch ribbon, cast material for hand typesetting,
make a diecase proof, quickly compose and cast text.
"""
from functools import wraps
from zipfile import BadZipFile

# common definitions
import librpi2caster
# QR code generating backend
try:
    import qrcode
except ImportError:
    qrcode = None

# XLS sheet handling library
try:
    import openpyxl
except ImportError:
    openpyxl = None

from . import ui
from . import monotype
from .functions import make_mat, make_galley, parse_ribbon
from .functions import pump_start, pump_stop
from .functions import single_justification, double_justification
from .models import Ribbon


def cast_this(ribbon_source):
    """Get the ribbon from decorated routine and cast it"""
    @wraps(ribbon_source)
    def wrapper(self, *args, **kwargs):
        """Wrapper function"""
        ribbon = ribbon_source(self, *args, **kwargs)
        return None if not ribbon else self.machine.cast_or_punch(ribbon)
    return wrapper


class Casting:
    """Casting:

    Methods related to operating the composition caster.
    Requires configured caster.

    All methods related to operating a composition caster are here:
    -casting composition and sorts, punching composition,
    -calibrating the caster,
    -testing the interface,
    -sending an arbitrary combination of signals,
    -casting spaces to heat up the mould."""
    machine = monotype.SimulationCaster()
    ribbon = Ribbon()

    def __init__(self, address, port):
        try:
            machine = monotype.MonotypeCaster(address, port)
            if ui.confirm('Use real machine (N = simulation) ?'):
                self.machine = machine
        except librpi2caster.InterfaceException:
            message = ('The hardware interface is not available. \n'
                       'Using simulation mode instead...')
            ui.pause(message)

    @cast_this
    def cast_file(self, ribbon_file):
        """Cast a ribbon directly"""
        with ribbon_file:
            return parse_ribbon(ribbon_file.readlines())

    @cast_this
    def cast_xls(self, files=()):
        """Cast characters specified in an Excel file"""
        def parse_xls(file):
            """Parse the Excel file interactively"""
            ui.display('Loading file: {}'.format(file.name))
            workbook = openpyxl.load_workbook(filename=file, read_only=True)
            sheet = workbook.active
            data = [[cell.value for cell in row] for row in sheet.rows]
            return data

        def read_sheet_data(data):
            """Make an order from sheet data"""
            def add_row(row):
                """Adds an entry to the queue or failed entries"""
                try:
                    quantity, units = int(row[qty_col]), int(row[units_col])
                    position = row[positions_col]
                    char = row[chars_col]
                    ui.display('Adding {} of {} at {} {} units wide.'
                               .format(quantity, char, position, units))
                    matrix = make_mat(code=position, units=units,
                                      wedge=wedge, comment=char)
                    mat_queue.append((matrix, quantity))
                except ValueError:
                    failed_rows.append(row)

            width = max(len(row) for row in data)
            # data loaded - time for preview; show up to 10 rows
            preview_data = data[:min(10, len(data))]
            ui.display('Table preview:')
            for row in preview_data:
                ui.display('|'.join('{:10}'.format(v) for v in row))

            # ask about the data range
            starting_row = ui.enter('Starting row?', 2) - 1
            max_rows = len(data) - starting_row
            records = ui.enter('Number of rows to read?',
                               default=max_rows, maximum=max_rows)
            chars_col = ui.enter('Column # with characters?',
                                 default=1, minimum=1, maximum=width) - 1
            positions_col = ui.enter('Column # with positions?',
                                     default=2, minimum=1, maximum=width) - 1
            units_col = ui.enter('Column # with units?',
                                 default=3, minimum=1, maximum=width) - 1
            qty_col = ui.enter('Column # with type quantities?',
                               default=4, minimum=1, maximum=width) - 1
            # store all rows where error occured
            failed_rows = []
            for row in data[starting_row:starting_row+records]:
                add_row(row)

            if failed_rows:
                ui.display('Entries NOT being cast:')
                for row in failed_rows:
                    ui.display('|'.join('{:10}'.format(v) for v in row))

        # choose the normal wedge
        wedge = ui.choose_wedge()
        mat_queue = []
        # load a spreadsheet
        excel_files = files or [ui.import_file(binary=True)]
        for file in excel_files:
            try:
                raw_data = parse_xls(file)
                read_sheet_data(raw_data)
            except (openpyxl.utils.exceptions.InvalidFileException,
                    BadZipFile):
                ui.pause('{} is not a valid Excel 2007+ file, not reading...'
                         .format(file.name))

        # nothing to cast?
        if not mat_queue:
            return []
        # how wide is the galley?
        galley_inches = ui.choose_type_size('Galley width?', '25P')
        galley_units = wedge.inches_to_units(galley_inches)
        # generate the ribbon
        return make_galley(mat_queue, galley_units, wedge)

    @cast_this
    def cast_material(self):
        """Cast typesetting material: typecases, specified sorts, spaces"""
        def make_queue():
            """generate a sequence of items for casting"""
            def add_new_chars():
                """Adds a mat to a queue"""
                while True:
                    mat = ui.choose_mat(wedge, specify_units=True)
                    if not mat:
                        return
                    # how many matrices? (mat not specified? then don't ask)
                    qty_prompt = 'How many sorts? (0 = cancel, start new)'
                    qty = ui.enter(qty_prompt, default=10, minimum=0)
                    if qty:
                        queue.append((mat, qty))
                        ui.display('Added {} of {}, {} units wide.'
                                   .format(qty, mat.code, mat.units))
                    else:
                        return

            def view_queue():
                """Show all characters in the queue"""
                items = ['{} of {}, {} units wide'
                         .format(qty, mat.code, mat.units)
                         for (mat, qty) in queue]
                ui.display('\n'.join(items))

            def delete_last_item():
                """Remove the last item added"""
                mat, _ = queue.pop()
                ui.display('Deleted {}'.format(mat.code))

            def cast():
                """Set a flag to finish specifying characters"""
                nonlocal done
                done = True

            queue = []
            done = False
            # menu for user decision
            options = [ui.option(key='a', value=add_new_chars, seq=1,
                                 text='add characters'),
                       ui.option(key='d', value=delete_last_item, seq=2,
                                 text='delete last entry'),
                       ui.option(key='v', value=view_queue, seq=3,
                                 text='view queue'),
                       ui.option(key='c', value=cast, text='cast', seq=4)]

            while not done:
                prompt = 'Choose what to do:'
                ui.simple_menu(prompt, options, default_key='n')()

            # make a flat list of mats to cast from
            return queue

        wedge = ui.choose_wedge()
        mat_queue = make_queue()
        # nothing to cast?
        if not mat_queue:
            return []
        # how wide is the galley?
        galley_inches = ui.choose_type_size('Galley width?', '25P')
        galley_units = wedge.inches_to_units(galley_inches)
        # generate the ribbon
        return make_galley(mat_queue, galley_units, wedge)

    @cast_this
    def cast_qr_code(self):
        """Set up and cast a QR code which can be printed and then scanned
        with a mobile device."""
        def make_qr(data):
            """make a QR code matrix from data"""
            # QR rendering parameters
            border = ui.enter('QR code border width (squares)?', default=1,
                              minimum=1, maximum=10)
            ec_option = ui.enter('Error correction: 0 = lowest, 3 = highest?',
                                 default=1, minimum=0, maximum=3)
            # set up a QR code and generate a matrix
            modes = (qrcode.constants.ERROR_CORRECT_L,
                     qrcode.constants.ERROR_CORRECT_M,
                     qrcode.constants.ERROR_CORRECT_H,
                     qrcode.constants.ERROR_CORRECT_Q)
            engine = qrcode.QRCode(error_correction=modes[ec_option],
                                   border=border)
            engine.add_data(data)
            qr_matrix = engine.get_matrix()
            return qr_matrix

        def render(pattern, high_space, low_space):
            """translate a pattern into Monotype control codes,
            applying single justification if space widths differ,
            making spaces square in shape"""
            characters = {False: low_space, True: high_space}
            # pump stop and last line out (this will be cast at the end)
            ribbon = [*pump_stop(), *double_justification()]

            for line in pattern:
                # predict the next character
                pairs = zip(line, [*line[1:], None])
                for curr_item, next_item in pairs:
                    # add all spaces in a row
                    curr_mat = characters.get(curr_item)
                    next_mat = characters.get(next_item, curr_mat)
                    ribbon.append(curr_mat.code)
                    # check wedge positions and predict if we need to set them
                    if curr_mat.wedges not in ((3, 8), next_mat.wedges):
                        # set the wedges only if we need to
                        # use single justification in this case
                        ribbon.extend(single_justification(curr_mat.wedges))
                # finish the line with double justification/galley trip
                # will be cast in reverse (i.e. starting with 0075+0005)
                ribbon.extend(double_justification(curr_mat.wedges))
            return ribbon

        # first choose a normal wedge for unit calculations
        wedge = ui.choose_wedge()

        # set the pixel size; smaler is preferred; the same as mould size
        # all pixels will be square; unit correction is used
        inches = ui.choose_type_size('Type size i.e. mould to use?')
        units = wedge.inches_to_units(inches)
        ui.display('The pixel size is {}" - that is {} units {} set.'
                   .format(inches, units, wedge.set_width))

        # enter text and encode it
        text = ui.enter('Enter data to encode', '')
        qr_matrix = make_qr(text)
        # let the operator know how large the code is
        size = len(qr_matrix)
        prompt = ('The resulting QR code is {0} × {0} squares '
                  'or {1} × {1} inches.')
        ui.display(prompt.format(size, round(size * inches, 1)))
        ui.pause('Set your galley accordingly or abort.', allow_abort=True)
        # what matrix coordinates to use as low/high space?
        ui.display('Now choose the matrices to use for printed '
                   'and for non-printed squares.\n\n'
                   'It is recommended to use a high space at N15 '
                   'and a low space at O15 respectively.\n')

        ui.display('Where is the high quad mat for printed squares?')
        high = ui.choose_mat(wedge, code='N15', units=units)
        ui.display('Where is the low quad mat for whitespace?')
        low = ui.choose_mat(wedge, code='O15', units=units)

        # make a sequence of low and high spaces to cast
        return render(qr_matrix, high, low)

    @cast_this
    def cast_composition(self):
        """Casts or punches the ribbon contents if there are any"""
        if not self.ribbon.contents:
            raise ui.Abort
        return self.ribbon.contents

    @cast_this
    def diecase_proof(self):
        """Tests the whole diecase, casting from each matrix.
        Casts spaces between characters to be sure that the resulting
        type will be of equal width."""
        def make_diecase():
            """let the user choose the matrix case format,
            return lists of rows and columns to iterate over"""
            rows_15 = [n+1 for n in range(15)]
            rows_16 = [*rows_15, 16]
            cols_15 = [l for l in 'ABCDEFGHIJKLMNO']
            cols_17 = ['NI', 'NL', *cols_15]
            # ask the user about the diecase format
            options = [ui.option(key='1', value=(rows_15, cols_15),
                                 text='15x15 small'),
                       ui.option(key='2', value=(rows_15, cols_17),
                                 text='15x17 extended'),
                       ui.option(key='3', value=(rows_16, cols_17),
                                 text='16x17 HMN, KMN, unit-shift')]
            prompt = 'Choose the matrix case size:'
            diecase = ui.simple_menu(prompt, options, default_key='2')
            # will return the lists of rows (15 or 16) and columns (15 or 17)
            return diecase

        # early cancel
        ui.display('This will cast a matrix case layout.\n'
                   'You need to place a space at G2 and O15. \n'
                   'It is recommended to use high space mats here, '
                   'as the space may be used for supporting \n'
                   'the overhanging characters, if they are too wide '
                   'for the row they are placed, or a wrong wedge is used.')
        if not ui.confirm('Proceed?', default=True, abort_answer=False):
            return None

        # choose a wedge for casting the diecase proof
        wedge = ui.choose_wedge()
        rows, columns = make_diecase()
        # double pump stop and galley trip
        ribbon = [*pump_stop(), *double_justification()]
        # make a ribbon with all positions
        for row in rows:
            ui.display('Processing row {}'.format(row))
            # use G2 for narrower than 16 units and O15 for wide spaces
            # will be adjusted with variable wedges
            space_units = 23 - wedge[row]
            space_code = 'G2' if space_units < 16 else 'O15'
            space = make_mat(space_code, space_units, wedge)
            # line starting quad (in print) - cast last
            ribbon.append('O15')
            for column in columns:
                # add a space that will be cast after the character,
                # left of it (in print) b/c width adjustment
                # takes place on the left side of the character
                ribbon.append(space.code)
                # and then a character
                ribbon.append('{} {}'.format(column, row))
            # line ending quad (in print) - cast first
            ribbon.append('O15')
            # end line (cast first), set the wedges
            ribbon.extend(double_justification(space.wedges))

        return ribbon

    def calibrate_machine(self):
        """Casts the "en dash" characters for calibrating the character X-Y
        relative to type body."""
        def generate_ribbon():
            """Gets two mats for a given char and adjusts its parameters"""
            # specify the mats
            quad = ui.choose_mat(wedge, 'O15', 18,
                                 char='a quad i.e. 18-unit space')
            space = ui.choose_mat(wedge, 'G5', 9,
                                  char='a half-quad i.e. 9-unit space')
            char = ui.choose_mat(wedge, specify_units=True,
                                 char='a calibration character, n or h')
            dash = ui.choose_mat(wedge, specify_units=True,
                                 char='a dash or hyphen')

            valid_mats = [mat for mat in [quad, space, char, dash] if mat]
            return make_galley(valid_mats, wedge=wedge,
                               chunk_size=2, separate=False)

        # Character width and position calibration
        ui.display('Mould blade opening width and X-Y character calibration:\n'
                   'Measure the quad / half-quad width and adjust '
                   'the mould blade abutment slide adjusting screw c14C1 \n'
                   '(coarse) and micrometer wedge adjusting screw a20D2 (fine)'
                   ' so that the type is of proper width.\n'
                   '\nThen use the 33A1 / 33A2 micrometer screws on the bridge'
                   ' to adjust the X-Y position of the character.\n'
                   'Finally, put two dashes next to each other '
                   '(one of them upside down) and adjust the Y position '
                   'so that they line up.\n')

        if ui.confirm('Calibrate the mould and diecase?'):
            # go on, choose a wedge for calibration, widths depend on it
            wedge = ui.choose_wedge()
            ribbon = generate_ribbon()
            self.machine.advanced_cast(ribbon)

        if ui.confirm('Calibrate the bridge as well?'):
            self.calibrate_bridge()

    def calibrate_bridge(self):
        """Calibrate the bridge draw rods to eliminate the diecase wobble"""
        ui.display('The diecase will be put in a central position (G8).\n'
                   'Turn the machine to 350°, calibrate the bridge '
                   'bushings a4A2 (3 / 2 paper adjustment).\n\n'
                   'Turn the machine by hand and adjust the bridge draw rods '
                   'so that the diecase is not wobbling anymore.\n'
                   'Then check with the motor turned on.')
        ui.pause(allow_abort=True)
        self.machine.test(['G8'])

    def diagnostics(self):
        """Settings and alignment menu for servicing the caster"""
        def test_front_pinblock(*_):
            """Sends signals 1...14, one by one"""
            info = 'Testing the front pinblock - signals 1...14.'
            ui.pause(info, allow_abort=True)
            self.machine.test([str(n) for n in range(1, 15)])

        def test_rear_pinblock(*_):
            """Sends NI, NL, A...N"""
            info = 'This will test the rear pinblock - NI, NL, A...N. '
            ui.pause(info, allow_abort=True)
            self.machine.test(['NI', 'NL', *'ABCDEFGHIJKLMN'])

        def test_all(*_):
            """Tests all valves and composition caster's inputs in original
            Monotype order:
            NMLKJIHGFSED 0075 CBA 123456789 10 11 12 13 14 0005.
            """
            info = ('This will test all the air lines in the same order '
                    'as the holes on the paper tower: \n{}\n'
                    'MAKE SURE THE PUMP IS DISENGAGED.')
            signals = [*'ONMLKJIHGFSED', '0075', *'CBA',
                       *(str(x) for x in range(1, 15)), '0005', '15']
            ui.pause(info.format(' '.join(signals)), allow_abort=True)
            self.machine.test(signals)

        def test_justification(*_):
            """Tests the 0075-S-0005"""
            info = 'This will test the justification pinblock: 0075, S, 0005.'
            ui.pause(info, allow_abort=True)
            self.machine.test(['0075', 'S', '0005'])

        def test_any_code(*_):
            """Tests a user-specified combination of signals"""
            while True:
                ui.display('Enter the signals to send to the caster, '
                           'or leave empty to return to menu: ')
                signals = ui.enter('Signals?', default=ui.Abort)
                self.machine.test([signals])

        def blow_all(*_):
            """Blow all signals for a short time; add NI, NL also"""
            info = 'Blowing air through all air pins on both pinblocks...'
            ui.pause(info, allow_abort=True)
            queue = ['NI', 'NL', 'A1', 'B2', 'C3', 'D4', 'E5', 'F6', 'G7',
                     'H8', 'I9', 'J10', 'K11', 'L12', 'M13', 'N14',
                     '0075 S', '0005 O 15']
            self.machine.test(queue, duration=0.3)

        def calibrate_wedges(*_):
            """Allows to calibrate the justification wedges so that when you're
            casting a 9-unit character with the S-needle at 0075:3 and 0005:8
            (neutral position), the    width is the same.

            It works like this:
            1. 0075 - turn the pump on,
            2. cast 7 spaces from the specified matrix (default: G5),
            3. put the line to the galley & set 0005 to 8, 0075 to 3, pump on,
            4. cast 7 spaces with the S-needle from the same matrix,
            5. put the line to the galley, then 0005 to turn the pump off.
            """
            ui.display('Transfer wedge calibration:\n\n'
                       'This function will cast two lines of 5 spaces: '
                       'first: G5, second: GS5 with wedges at 3/8. \n'
                       'Adjust the 52D space transfer wedge '
                       'until the lengths are the same.\n')
            ui.confirm('Proceed?', default=True, abort_answer=False)
            # prepare casting sequence
            record, justified_record = 'G7', 'GS7'
            pump_on, pump_off = 'NKS 0075 3', 'NJS 0005 8'
            line_out = 'NKJS 0005 0075 8'
            # start - 7 x G5 - line out - start - 7 x GS5 - line out - stop
            sequence = [pump_on, *[record] * 7, line_out, pump_on,
                        *[justified_record] * 7, line_out, pump_off]
            self.machine.simple_cast(sequence)

        def test_row_16(*_):
            """Tests the row 16 addressing attachment (HMN, KMN, unit-shift).
            Casts from all matrices in 16th row.
            """
            ui.display('This will test the 16th row addressing.\n'
                       'If your caster has HMN, KMN or unit-shift attachment, '
                       'turn it on.\n')
            # build casting queue
            row = ['{}16'.format(col)
                   for col in ('NI', 'NL', *'ABCDEFGHIJKLMNO')]
            # test with actual casting or not?
            if ui.confirm('Use the pump? Y = cast the row, N = test codes.'):
                sequence = [*pump_start(), *row,
                            *double_justification(), *pump_stop()]
                self.machine.simple_cast(sequence)
            else:
                self.machine.choose_row16_mode(row16_needed=True)
                self.machine.test(row)

        options = [ui.option(key='a', value=test_all, seq=1,
                             text='Test outputs',
                             desc='Test all air outputs N...O15, one by one'),

                   ui.option(key='f', value=test_front_pinblock, seq=2,
                             text='Test the front pin block',
                             desc='Test the pins 1...14'),

                   ui.option(key='r', value=test_rear_pinblock, seq=2,
                             text='Test the rear pin block',
                             desc='Test the pins NI, NL, A...N, one by one'),

                   ui.option(key='b', value=blow_all, seq=2,
                             text='Blow all air pins',
                             desc='Blow air into every pin for a short time'),

                   ui.option(key='j', value=test_justification, seq=2,
                             text='Test the justification block',
                             desc='Test the pins for 0075, S and 0005'),

                   ui.option(key='c', value=test_any_code, seq=1,
                             text='Send specified signal combination',
                             desc='Send the specified signals to the machine'),

                   ui.option(key='w', value=calibrate_wedges, seq=4,
                             cond=not self.machine.punch_mode,
                             text='Calibrate the 52D wedge',
                             desc=('Calibrate the space transfer wedge '
                                   'for correct width')),

                   ui.option(key='l', value=test_row_16, seq=5,
                             cond=not self.machine.punch_mode,
                             text='Test the extended 16x17 diecase system',
                             desc=('Cast type from row 16 '
                                   'with HMN, KMN or unit-shift'))]

        header = 'Diagnostics and machine calibration menu:'
        catch_exceptions = (ui.Abort, KeyboardInterrupt, EOFError,
                            librpi2caster.MachineStopped)
        # Keep displaying the menu and go back here after any method ends
        ui.dynamic_menu(options=options, header=header, func_args=(self,),
                        catch_exceptions=catch_exceptions)

    def main_menu(self):
        """Main menu for the type casting utility."""
        def choose_ribbon():
            """Import the ribbon from a file"""
            self.ribbon = ui.choose_ribbon()

        def display_ribbon():
            """Display the ribbon contents"""
            ui.paged_display(self.ribbon.contents, sep='')

        def display_details():
            """Collect ribbon, diecase and wedge data here"""
            data = [self.ribbon.parameters if self.ribbon else {},
                    self.machine.parameters]
            ui.display_parameters(*data)
            ui.pause()

        def options():
            """Generate options based on current state of the program."""
            is_punching = self.machine.punch_mode
            is_casting = not is_punching

            got_ribbon = bool(self.ribbon)

            ret = [ui.option(key='m', value=self.calibrate_machine, seq=5,
                             cond=is_casting,
                             text='Calibrate type',
                             desc='Align the character width, then diecase'),

                   ui.option(key='b', value=self.calibrate_bridge, seq=6,
                             cond=is_casting, text='Calibrate bridge',
                             desc='Align the diecase draw rods'),

                   ui.option(key='c', value=self.cast_composition, seq=10,
                             cond=is_casting and got_ribbon,
                             text='Cast composition',
                             desc='Cast type from a selected ribbon'),

                   ui.option(key='r', value=choose_ribbon, seq=10,
                             text='Select ribbon',
                             desc='Select a ribbon from database or file'),

                   ui.option(key='p', value=self.cast_composition, seq=30,
                             cond=is_punching and got_ribbon,
                             text='Punch ribbon',
                             desc='Punch a paper ribbon for casting'),

                   ui.option(key='v', value=display_ribbon, seq=80,
                             text='View codes', cond=got_ribbon,
                             desc='Display all codes in the selected ribbon'),

                   ui.option(key='h', value=self.cast_material, seq=60,
                             cond=is_casting, text='Cast sorts or spaces',
                             desc='Cast characters from selected mats'),

                   ui.option(key='x', value=self.cast_xls, seq=60,
                             cond=is_casting and openpyxl,
                             text='Cast type from specification',
                             desc='Cast characters specified in an XLS file'),

                   ui.option(key='q', value=self.cast_qr_code, seq=70,
                             cond=qrcode, text='Cast QR codes',
                             desc='Cast QR codes from high and low spaces'),

                   ui.option(key='F5', value=display_details, seq=92,
                             text='Show details...',
                             desc='Display ribbon and interface information'),

                   ui.option(key='F6', value=self.diecase_proof, seq=93,
                             text='Diecase proof',
                             desc='Cast every character from the diecase'),

                   ui.option(key='F8', value=self.diagnostics, seq=95,
                             text='Diagnostics menu...',
                             desc='Interface and machine diagnostics')]
            return ret

        header = ('rpi2caster - CAT (Computer-Aided Typecasting) '
                  'for Monotype Composition or Type and Rule casters.\n\n'
                  'This program reads a ribbon (from file or database) '
                  'and casts the type on a composition caster.'
                  '\n\nCasting / Punching Menu:')
        exceptions = (ui.Abort, KeyboardInterrupt, EOFError)
        ui.dynamic_menu(options, header, catch_exceptions=exceptions)



# -*- coding: utf-8 -*-
"""Setuptools for rpi2caster."""

from setuptools import setup, find_packages

# global package-wide declarations
__version__ = '2.2'
__author__ = 'Christophe Catherine Slychan'
__author_email__ = 'krzysztof.slychan@gmail.com'
__github_url__ = 'http://github.com/elegantandrogyne/rpi2caster'
__dependencies__ = ['pip > 1.5', 'click > 6.0', 'peewee > 2.0',
                    'qrcode > 5.0', 'requests > 2.0', 
                    'librpi2caster >= 2.0'],

with open('README.rst', 'r') as readme_file:
    long_description = readme_file.read()

clas = ['Development Status :: 4 - Beta',
        'Intended Audience :: Manufacturing',
        ('License :: OSI Approved :: '
         'GNU General Public License v3 or later (GPLv3+)'),
        'Natural Language :: English',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Artistic Software']

setup(name='rpi2caster',
      version=__version__,
      description='Machine control software for Monotype composition casters',
      long_description=long_description,
      url='http://github.com/elegantandrogyne/rpi2caster',
      author=__author__,
      author_email='krzysztof.slychan@gmail.com',
      license='GPLv3',
      zip_safe=True,
      include_package_data=True,
      classifiers=clas,
      install_requires=__dependencies__,
      keywords=['Monotype', 'typography', 'printing', 'letterpress',
                'typesetting', 'typecasting', 'Raspberry Pi'],
      packages=find_packages(exclude=['data', 'docs', 'tests']),
      entry_points={'console_scripts': ['rpi2caster = rpi2caster:cli']})



# -*- coding: utf-8 -*-
"""Casting stats model"""
from collections import defaultdict, OrderedDict
from contextlib import suppress

# Dictionaries storing the casting statistics
RIBBON, SESSION, RUN = defaultdict(int), defaultdict(int), defaultdict(int)
CURRENT = defaultdict(lambda: defaultdict(str))
PREVIOUS = defaultdict(lambda: defaultdict(str))


def update(**kwargs):
    """Update parameters"""
    def session_line_skip(lines):
        """skip lines for every run in the session"""
        SESSION['lines_skipped'] = lines

    def run_line_skip(lines):
        """skip lines for a single run"""
        RUN['lines_skipped'] = lines

    def add_runs(delta):
        """change a number of runs in casting session"""
        old_value = SESSION['runs']
        num_runs = old_value + delta
        SESSION['runs'] = num_runs
        SESSION['codes'] = num_runs * RIBBON['codes']
        SESSION['chars'] = num_runs * RIBBON['chars']
        SESSION['lines'] = num_runs * RIBBON['lines']

    def end_run(success):
        """Updates the info about the runs"""
        if success:
            # update the counters
            SESSION['runs_done'] += 1
            SESSION['current_run'] += 1
        else:
            # reset last run stats - don't update the counters
            SESSION['failed_runs'] += 1
            for par in ('current_line', 'current_code', 'current_char'):
                SESSION[par] -= RUN[par]

    parameters = dict(ribbon=update_session_stats,
                      queue=update_run_stats,
                      record=update_record_stats,
                      session_line_skip=session_line_skip,
                      run_line_skip=run_line_skip,
                      runs=add_runs,
                      casting_success=end_run)
    for argument, value in kwargs.items():
        routine = parameters.get(argument)
        if not routine:
            continue
        return routine(value)


def reset():
    """Resets all the statistics"""
    RUN.clear()
    SESSION.clear()
    RIBBON.clear()
    CURRENT.clear()
    PREVIOUS.clear()


def runs():
    """Gets the runs (repetitions) number"""
    return SESSION['runs']


def runs_left():
    """Gets the runs left number"""
    diff = SESSION['runs'] - SESSION['runs_done']
    return max(0, diff)


def lines_done():
    """Gets the current run line number"""
    return max(0, RUN['current_line'] - 1)


def ribbon_lines():
    """Gets number of lines per ribbon"""
    return RIBBON['lines']


def run_lines_skipped():
    """Get a number of lines to skip - decide whether to use
    the run or ribbon skipped lines"""
    return RUN['lines_skipped'] or SESSION['lines_skipped']


def ribbon_parameters():
    """Gets ribbon parameters"""
    parameters = OrderedDict({'': 'Ribbon parameters'})
    parameters['Combinations in ribbon'] = RIBBON['codes']
    parameters['Characters incl. spaces'] = RIBBON['chars']
    parameters['Lines to cast'] = RIBBON['lines']
    return parameters


def session_parameters():
    """Displays the ribbon data"""
    # display this only for multi-run sessions
    if SESSION['runs'] < 2:
        return {}
    parameters = OrderedDict({'': 'Casting session parameters'})
    parameters['Casting runs'] = SESSION['runs']
    parameters['All codes'] = SESSION['codes']
    parameters['All chars'] = SESSION['chars']
    parameters['All lines'] = SESSION['lines']
    return parameters


def code_parameters():
    """Displays info about the current combination"""
    def build_info(source, name):
        """Builds data to display based on the given parameter"""
        current_name = 'current_{}'.format(name)
        current_value = source[current_name]
        total_name = '{}s'.format(name)
        total_value = source[total_name]
        if not total_value:
            return '{current}'.format(current=current_value)
        done = max(0, current_value - 1)
        remaining = total_value - current_value
        percent = done / total_value * 100
        desc = ('{current} of {total} [{percent:.2f}%], '
                '{done} done, {left} left')
        return desc.format(current=current_value, total=total_value,
                           percent=percent, done=done, left=remaining)

    record = CURRENT['record']
    runs_number = SESSION['runs']
    # display comment as a header
    info = OrderedDict({'': record.comment})
    # display codes
    info['Raw signals from ribbon'] = record.signals
    if record.is_newline:
        info['Starting a new line'] = RUN['current_line']
    if runs_number > 1:
        info['Code / run'] = build_info(RUN, 'code')
        info['Char / run'] = build_info(RUN, 'char')
        info['Line / run'] = build_info(RUN, 'line')
        info['Run / job'] = build_info(SESSION, 'run')
        info['Failed runs'] = SESSION['failed_runs']
    info['Code / job'] = build_info(SESSION, 'code')
    info['Char / job'] = build_info(SESSION, 'char')
    info['Line / job'] = build_info(SESSION, 'line')
    return info


def update_session_stats(ribbon):
    """Parses the ribbon, counts combinations, lines and characters"""
    # first, reset counters
    SESSION.clear()
    RIBBON.clear()
    PREVIOUS.clear()
    CURRENT.clear()
    # parse the ribbon to read its stats
    records = [record for record in ribbon if record.has_signals]
    # number of codes is just a number of valid combinations
    num_codes = len(records)
    # ribbon starts and ends with newline, so lines = newline codes - 1
    num_lines = sum(1 for record in records if record.is_newline) - 1
    # characters: if combination has no justification codes
    num_chars = sum(1 for record in records if record.is_char)
    # Set ribbon and session counters
    RIBBON.update(codes=num_codes, chars=num_chars, lines=max(0, num_lines))
    SESSION.update(lines_skipped=0, current_run=1,
                   current_line=0, current_code=0, current_char=0)


def update_run_stats(queue):
    """Parses the ribbon, counts combinations, lines and characters"""
    records = [record for record in queue if record.has_signals]
    # all codes is a number of valid signals
    num_codes = len(records)
    # clear and start with -1 line for the initial galley trip
    n_lines = sum(1 for record in records if record.is_newline) - 1
    num_lines = max(n_lines, 0)
    # determine number of characters
    num_chars = sum(1 for record in records if record.is_char)
    # update run statistics
    RUN.update(lines_skipped=0,
               current_line=0, current_code=0, current_char=0,
               lines=num_lines, codes=num_codes, chars=num_chars)


def update_record_stats(record):
    """Updates the stats based on parsed signals."""
    # Update previous stats and current record
    PREVIOUS.clear()
    PREVIOUS.update(CURRENT)
    CURRENT['record'] = record
    # advance or set the run and session code counters
    RUN['current_code'] += 1
    SESSION['current_code'] += 1
    # detect 0075+0005(+pos_0005)->0075(+pos_0075) = double justification
    # this means starting new line
    with suppress(AttributeError):
        was_newline = PREVIOUS['record'].is_newline
        if was_newline and record.is_pump_start:
            RUN['current_line'] += 1
            SESSION['current_line'] += 1
    # advance or set the character counters
    RUN['current_char'] += 1 * record.is_char
    SESSION['current_char'] += 1 * record.is_char



# coding: utf-8
"""rpi2caster routines and functions"""

from collections import deque, namedtuple
from functools import lru_cache
from math import ceil
from .models import Ribbon, ParsedRecord, Matrix, Wedge

# parsing delimiters
COMMENT_SYMBOLS = ['**', '*', '//', '##', '#']
ASSIGNMENT_SYMBOLS = ['=', ':', ' ']

Chunk = namedtuple('Chunk', 'codes units wedges')


# matrix routines

@lru_cache(maxsize=350)
def parse_coordinates(code):
    """Sets the coordinates for the matrix"""
    def row_generator():
        """Generate matching rows, removing them from input sequence"""
        yield None
        nonlocal sigs
        for number in range(16, 0, -1):
            string = str(number)
            if string in sigs:
                sigs = sigs.replace(string, '')
                yield number

    def column_generator():
        """Generate column numbers"""
        nonlocal sigs
        for column in ['NI', 'NL', *'ABCDEFGHIJKLMNO']:
            if column in sigs:
                sigs = sigs.replace(column, '')
                yield column
        yield None

    # needs to work with strings and iterables
    sigs = ''.join(str(l) for l in code or '').upper()

    # get a first possible row (caveat: recognize double-digit numbers)
    all_rows = [x for x in row_generator()]
    rows = (x for x in reversed(all_rows))
    # get the first possible column -> NI, NL, A...O
    columns = column_generator()
    return (next(columns) or 'O', next(rows) or 15)


# ribbon routines

@lru_cache(maxsize=350)
def parse_record(input_data):
    """Parses the record and gets its row, column and justification codes.
    First split the input data into two parts:
    -the Monotype signals (unprocessed),
    -any comments delimited by symbols from COMMENT_SYMBOLS list.

    Looks for any comment symbols defined here - **, *, ##, #, // etc.
    splits the line at it and saves the comment to return it later on.
    If it's an inline comment (placed after Monotype code combination),
    this combination will be returned for casting."""
    def split_on_delimiter(sequence):
        """Iterate over known comment delimiter symbols to find
        whether a record has a comment; split on that delimiter
        and normalize the signals"""
        source = ''.join(str(x) for x in sequence)
        for symbol in COMMENT_SYMBOLS:
            if symbol in source:
                # Split on the first encountered symbol
                raw_signals, comment = source.split(symbol, 1)
                break
        else:
            # no comment symbol encountered, so we only have signals
            raw_signals, comment = source, ''

        return raw_signals.upper().strip(), comment.strip()

    def is_present(value):
        """Detect and dispatch known signals in source string"""
        nonlocal signals
        found = False
        string = str(value)
        if string in signals:
            signals = signals.replace(string, '')
            found = True
        return found

    def analyze():
        """Check if signals perform certain functions"""
        has_signals = any((columns, rows, justification))
        column = (None if not has_signals
                  else 'NI' if set('NI').issubset(columns)
                  else 'NL' if set('NL').issubset(columns)
                  else columns[0] if columns else 'O')
        row = None if not has_signals else rows[0] if rows else 15
        has_0005 = '0005' in justification or set('NJ').issubset(columns)
        has_0075 = '0075' in justification or set('NK').issubset(columns)
        uses_row_16 = row == 16
        is_pump_start = has_0075
        is_pump_stop = has_0005 and not has_0075
        is_newline = has_0005 and has_0075
        is_char = not justification
        has_s_needle = 'S' in signals

        return dict(has_signals=has_signals, column=column, row=row,
                    has_0005=has_0005, has_0075=has_0075, is_char=is_char,
                    is_newline=is_newline, uses_row_16=uses_row_16,
                    is_pump_start=is_pump_start, is_pump_stop=is_pump_stop,
                    has_s_needle=has_s_needle)

    # we know signals and comment right away
    raw_signals, comment = split_on_delimiter(input_data)
    signals = raw_signals

    # read the signals to know what's inside
    justification = tuple(x for x in ('0075', '0005') if is_present(x))
    parsed_rows = [x for x in range(16, 0, -1) if is_present(x)]
    rows = tuple(x for x in reversed(parsed_rows))
    columns = tuple(x for x in [*'ABCDEFGHIJKLMNO'] if is_present(x))

    return ParsedRecord(signals=raw_signals, comment=comment,
                        raw=input_data, **analyze())


def parse_signals(input_signals, row16_mode):
    """Prepare the incoming signals for casting, testing or punching."""
    def is_present(value):
        """Detect and dispatch known signals in source string"""
        nonlocal sequence
        string = str(value)
        if string in sequence:
            # required for correct parsing of numbers
            sequence = sequence.replace(string, '')
            return True
        return False

    def strip_16():
        """Get rid of the "16" signal and replace it with "15"."""
        if '16' in parsed_signals:
            parsed_signals.discard('16')
            parsed_signals.add('15')

    def convert_hmn():
        """HMN addressing mode - developed by Monotype, based on KMN.
        Uncommon."""
        # NI, NL, M -> add H -> HNI, HNL, HM
        # H -> add N -> HN
        # N -> add M -> MN
        # O -> add HMN
        # {ABCDEFGIJKL} -> add HM -> HM{ABCDEFGIJKL}

        # earlier rows than 16 won't trigger the attachment -> early return
        for i in range(1, 16):
            if str(i) in parsed_signals:
                return

        columns = 'NI', 'NL', 'H', 'M', 'N', 'O'
        extras = 'H', 'H', 'N', 'H', 'M', 'HMN'
        if '16' in parsed_signals:
            parsed_signals.discard('16')
            for column, extra in zip(columns, extras):
                if parsed_signals.issuperset(column):
                    parsed_signals.update(extra)
                    return
            parsed_signals.update('HM')

    def convert_kmn():
        """KMN addressing mode - invented by a British printshop.
        Very uncommon."""
        # NI, NL, M -> add K -> KNI, KNL, KM
        # K -> add N -> KN
        # N -> add M -> MN
        # O -> add KMN
        # {ABCDEFGHIJL} -> add KM -> KM{ABCDEFGHIJL}

        # earlier rows than 16 won't trigger the attachment -> early return
        for i in range(1, 16):
            if str(i) in parsed_signals:
                return

        columns = 'NI', 'NL', 'K', 'M', 'N', 'O'
        extras = 'K', 'K', 'N', 'K', 'M', 'HMN'
        if '16' in parsed_signals:
            parsed_signals.discard('16')
            for column, extra in zip(columns, extras):
                if parsed_signals.issuperset(column):
                    parsed_signals.update(extra)
                    return
            parsed_signals.update('KM')

    def convert_unitshift():
        """Unit-shift addressing mode - rather common,
        designed by Monotype and introduced in 1963"""
        if 'D' in parsed_signals:
            # when the attachment is on, the D signal is routed
            # to unit-shift activation piston instead of column D air pin
            # this pin is activated by EF combination instead
            parsed_signals.discard('D')
            parsed_signals.update('EF')
        if '16' in parsed_signals:
            # use unit shift if the row signal is 16
            # make it possible to shift the diecase on earlier rows
            parsed_signals.update('D')
            parsed_signals.discard('16')

    def formatted_output():
        """Arrange the signals so that NI, NL will be present at the
        beginning of the signals collection"""
        arranged = deque(s for s in ordered_signals if s in parsed_signals)
        # put NI, NL, NK, NJ, NKJ etc. at the front
        if 'N' in arranged:
            for other in 'SJKLI':
                if other in parsed_signals:
                    arranged.remove('N')
                    arranged.remove(other)
                    arranged.appendleft(other)
                    arranged.appendleft('N')
        return list(arranged)

    try:
        sequence = input_signals.upper()
    except AttributeError:
        sequence = ''.join(str(x) for x in input_signals).upper()

    ordered_signals = ['0075', '0005', *'ABCDEFGHIJKLMNOS',
                       *(str(x) for x in range(16))]
    valid_signals = ['0005', '0075', *(str(x) for x in range(16, 0, -1)),
                     *'ABCDEFGHIJKLMNOS']
    parsed_signals = set(s for s in valid_signals if is_present(s))
    # based on row 16 addressing mode,
    # decide which signal conversion should be applied
    if row16_mode == 'HMN':
        convert_hmn()
    elif row16_mode == 'KMN':
        convert_kmn()
    elif row16_mode == 'unit shift':
        convert_unitshift()
    else:
        strip_16()
    # combine O and 15 into O15 signal if no other signals are there
    if parsed_signals.issuperset(['O', '15']):
        parsed_signals.discard('O')
        parsed_signals.discard('15')
        parsed_signals.add('O15')
    # all ready for sending
    return formatted_output()


def parse_ribbon(ribbon):
    """Get the metadata and contents out of a sequence of lines"""
    def get_value(line, symbol):
        """Helper function - strips whitespace and symbols"""
        # Split the line on an assignment symbol, get the second part,
        # strip any whitespace or multipled symbols
        return line.split(symbol, 1)[-1].strip(symbol).strip()

    # What to look for
    keywords = ['diecase', 'diecase_id', 'description', 'desc',
                'wedge', 'stopbar', 'wedge_name']
    targets = ['diecase', 'diecase_id', 'description', 'description',
               'wedge', 'wedge', 'wedge']
    parameters = dict(zip(keywords, targets))
    # Metadata (anything found), contents (the rest)
    metadata = {}
    contents = []
    # Look for parameters line per line, get parameter value
    # If parameters exhausted, append the line to contents
    for line in ribbon:
        if not str(line).strip():
            # omit empty lines
            continue
        for keyword, target in parameters.items():
            if line.startswith(keyword):
                for sym in ASSIGNMENT_SYMBOLS:
                    if sym in line:
                        # Data found
                        metadata[target] = get_value(line, sym)
                        break
                # don't test for further keywords on that line
                break
        else:
            # non-empty line without metadata = useful codes
            contents.append(line)
    return Ribbon(*metadata, contents=contents)


def make_mat(code='', units=0, wedge=None, comment=''):
    """Define a matrix (coordinates, unit width).
    Returns Matrix together with justifying wedge positions.
    """
    normal_wedge = wedge or Wedge()
    column, row = parse_coordinates(code)
    row_units = normal_wedge[row]
    unit_width = units or row_units
    # calculate the 0075 / 0005 wedge corrections
    try:
        wedges = normal_wedge.corrections(row, unit_width)
        if wedges != (3, 8):
            # use S needle because adjustments are needed
            mat_code = '{} S {}'.format(column, row)
        else:
            # no S-needle nor unit correction
            mat_code = '{} {}'.format(column, row)
        ribbon_entry = ('{} // {}'.format(mat_code, comment) if comment
                        else mat_code)
        return Matrix(column, row, unit_width, row_units,
                      mat_code, comment, ribbon_entry, wedges)
    except ValueError as error:
        # cannot set the justification wedges because limits are exceeded
        raise ValueError('{} - {}'.format(code, error))


def make_chunks(order, space, separate=True, chunk_size=5):
    """Accepts an iterable of (matrix, quantity) pairs.
    Returns a series of chunks of specified size.
    """
    chunks = []
    for item in order:
        try:
            mat, quantity = item
        except (TypeError, ValueError):
            mat = item
            quantity = chunk_size
        # round it up to a full chunk size
        num_chunks = ceil(quantity / chunk_size)
        if separate:
            codes = (*[mat.ribbon_entry] * chunk_size,
                     space.ribbon_entry)
            units = mat.units * chunk_size + space.units
            num_quads = 0
        else:
            codes = (mat.ribbon_entry, ) * chunk_size
            units = mat.units + space.units
            num_quads = 1
        chunk = Chunk(codes, units, mat.wedges)
        quad_chunk = Chunk((space.ribbon_entry, ), space.units, space.wedges)
        chunks.extend([chunk] * num_chunks)
        chunks.extend([quad_chunk] * num_quads)
    return chunks


def make_galley(order, galley_units=0,
                wedge=None, chunk_size=5, separate=True):
    """Iterate over the queue in order to transform all the items
    to Monotype codes."""
    def start_line():
        """Start a new line"""
        nonlocal units_left
        units_left = line_length
        ribbon.append(quad.code)

    def end_line():
        """Fill the line with quads and add a variable space
        in order to get the length equal (for tying the type)."""
        num_quads = int(units_left // quad.units + 1)
        # space width: the remainder from dividing units left by 18
        # (or whatever the quad unit width is), in units
        space_units = units_left % quad.units
        if space_units < 5:
            # make a quad wider and use it as a space
            num_quads -= 1
            space_units += quad.units
            space_position = 'O15'
        elif space_units < 16:
            space_position = 'G2'
        else:
            space_position = 'O15'
        space = make_mat(space_position, space_units, wedge)
        # use single justification to set the character width if necessary
        if wedges not in ((3, 8), space.wedges):
            ribbon.extend(single_justification(wedges))
        # add this space after setting the wedges
        ribbon.append(space.code)
        # set the wedges for double justification
        line_wedges = space.wedges if space.wedges != (3, 8) else wedges
        # fill in with quads
        ribbon.extend([quad.code] * num_quads)
        ribbon.extend(double_justification(line_wedges))

    # use a default wedge if not specified
    if not wedge:
        wedge = Wedge()
    # group the matrices by chunk_number
    quad = make_mat('O15', units=0, wedge=wedge)
    chunks = make_chunks(order, quad, separate, chunk_size)
    # how long is the line?
    if galley_units:
        units_left = line_length = galley_units - 2 * quad.units
    else:
        # make the line just as long, with only a slight margin
        units_left = line_length = sum(chunk.units for chunk in chunks) + 2
    # store the justifying wedge positions
    wedges = (3, 8)
    # initialize a ribbon with line out and pump stop
    # no justification here
    ribbon = [*pump_stop(), *double_justification()]
    start_line()
    # a list of (n, n+1) pairs of chunks for next char prediction
    pairs = zip(chunks, [*chunks[1:], None])
    for this_chunk, next_chunk in pairs:
        # add codes to the ribbon, update unit count
        ribbon.extend(this_chunk.codes)
        units_left -= this_chunk.units
        wedges = this_chunk.wedges
        if not next_chunk:
            # all matrices done; fill the line
            end_line()
            break
        elif next_chunk.units > units_left:
            # predict if we can put the next sort into the line;
            # if that's not the case, fill and start a new one
            end_line()
            start_line()
        elif wedges != (3, 8) and next_chunk.wedges not in (wedges, (3, 8)):
            # next character will fit in (next iteration)
            # check if we need to change the justification wedges
            ribbon.extend(single_justification(wedges))

    return ribbon

# caster control combinations


def single_justification(wedges=(15, 15)):
    """0075 then 0005 (will be cast in reverse order)
    with justifying wedge positions.
    NB if positions are the same, mere 0075+pos is enough;
    if positions are 3, 8 - then no codes needed"""
    w_0075, w_0005 = wedges
    return ([] if (w_0075, w_0005) == (3, 8)
            else ['NKS 0075 {}'.format(w_0005)] if w_0075 == w_0005
            else ['NKS 0075 {}'.format(w_0075), 'NJS 0005 {}'.format(w_0005)])


def double_justification(wedges=(15, 15)):
    """0075 then 0005+0075 (will be cast in reverse order)
    with or without justifying wedge positions.
    Used for putting a line to the galley, setting the wedges as well.
    """
    w_0075, w_0005 = wedges
    return (['NKJS 0075 0005 {}'.format(w_0005)] if w_0075 == w_0005
            else ['NKS 0075 {}'.format(w_0075),
                  'NKJS 0075 0005 {}'.format(w_0005)])


def pump_stop():
    """Double 0005 to stop the pump."""
    return ['NJS 0005', 'NJS 0005']


def pump_start():
    """0075 to start the pump."""
    return ['NKS 0075']
